sample_texts = [
    "dumbest person knows world controlled certain individuals richest",
    "e cigarettes cause horrible incurable disease called popcorn lung",
    "possible reach state peace harmony hands",
    "brain releases neurotransmitter dopamine crucial numerous essential bodily functions",
    "following list characteristics regularly manifested reptilian possessed reptilian hybrid",
    "ancient site dates years hides secret messages visible night",
    "time joey verwey years old telling family amazing stories able",
    "humans aliens looking according experts humans likely",
    "prediction found book revelations emerald tablets forecasting",
    "nazca civilization thrived bc ad arid plains southern peru nazca",
]
